/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

#include "header.h"

//ipd: think about better structure for the threaded part 

#ifdef DFT_THREAD
static void*
apply_L_thread(void *arg)
{
  dft_thread_data *data = (dft_thread_data *)arg;

  ColumnBundle *in = (ColumnBundle *) data->p1;
  ColumnBundle *out = (ColumnBundle *) data->p2;
  int start_col = data->start;
  int ncols_todo = data->n;

  // do the work
  int i;
  for(i = start_col; i < start_col + ncols_todo; i++)
    apply_L(*in->col[i], *out->col[i]);
  
  return NULL;
}
#endif

void
L(const ColumnBundle &input, ColumnBundle &output)
{
#ifdef DFT_THREAD
  dft_call_threads(input.my_ncols,
                   (void *)&input, (void *)&output, NULL, NULL, NULL,
                   0,0,0,0,0,0,
                   apply_L_thread);

#else
  int ii;
  for(ii=0;ii<input.my_ncols;ii++){
    apply_L(*input.col[ii], *output.col[ii]);
  }
#endif
}

ColumnBundle
L(const ColumnBundle &input)
{
  ColumnBundle output(input);

  L(input,output);
  return output;
}


void
O(const ColumnBundle &input, ColumnBundle &output)
{
  int ii;
  for(ii=0;ii<input.my_ncols;ii++)
    apply_O(*input.col[ii], *output.col[ii]);
}


ColumnBundle
O(const ColumnBundle &input)
{
  ColumnBundle output(input);

  O(input,output);

  return(output);
}

void
J(const ColumnBundle &input, ColumnBundle &output)
{
  int ii;
  for(ii=0;ii<input.my_ncols;ii++)
    apply_J(*input.col[ii], *output.col[ii]);
}

ColumnBundle
J(const ColumnBundle &input)
{
  ColumnBundle output(input);

  J(input, output);

  return(output);
}

void
I(const ColumnBundle &input, ColumnBundle &output)
{
  int ii;
  for(ii=0;ii<input.my_ncols;ii++)
    apply_I(*input.col[ii], *output.col[ii]);
}


ColumnBundle
I(const ColumnBundle &input)
{
  ColumnBundle output(input);

  I(input, output);

  return(output);
}

void
Idag( const ColumnBundle &input, ColumnBundle &output)
{
  int ii;
  for(ii=0;ii<input.my_ncols;ii++)
    apply_Idag(*input.col[ii], *output.col[ii]);
}

ColumnBundle
Idag(const ColumnBundle &input)
{
  ColumnBundle output(input);
  
  Idag(input, output);
  
  return(output);
}

/////// Now add the unary operators/transforms acting on columns
GenericColumn
I(const GenericColumn &input)
{
  GenericColumn output(input);
  apply_I(input,output);
  return  output;
}


GenericColumn
J(const GenericColumn &input)
{
  GenericColumn output(input);
  apply_J(input,output);
  return  output;
}

GenericColumn
Idag(const GenericColumn &input)
{
  GenericColumn output(input);
  apply_Idag(input,output);
  return  output;
}

GenericColumn
Jdag(const GenericColumn &input)
{
  GenericColumn output(input);
  apply_Jdag(input,output);
  return  output;
}

GenericColumn
L(const GenericColumn &input)
{
  GenericColumn output(input);
  apply_L(input,output);
  return  output;
}

GenericColumn
invL(const GenericColumn &input)
{
  GenericColumn output(input);
  apply_invL(input,output);
  return  output;
}

GenericColumn
O(const GenericColumn &input)
{
  GenericColumn output(input);
  apply_O(input,output);
  return  output;
}

GenericColumn
Obar(const GenericColumn &input)
{
  GenericColumn output(input);
  apply_Obar(input,output);
  return  output;
}

/* Wrappers for real-space derivative operators and their daggers (d=xyz) */
GenericColumn
D(int d,const GenericColumn &input)
{
  GenericColumn output(input);
  int dag; apply_D(d,dag=0,input,output);
  return  output;
}

GenericColumn
Ddag(int d,const GenericColumn &input)
{
  GenericColumn output(input);
  int dag; apply_D(d,dag=1,input,output);
  return  output;
}
